from __future__ import print_function
import json
import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize

import dpx_control.dpx_settings as ds

class DPX_test_pulse(object):
    def testPulseInit(self, slot, column=0):
        # Set Polarity to hole and DosiMode in OMR
        OMRCode = self.OMR

        if not isinstance(self.OMR, basestring):
            OMRCode[0] = 'DosiMode'
        else:
            OMRCode = '%04x' % ((int(self.OMR, 16) & ~((0b11) << 22)) | (0b10 << 22))
        
        self.DPXWriteOMRCommand(slot, OMRCode)

        if column == 'all':
            columnRange = range(16)
        elif not isinstance(column, basestring):
            if isinstance(column, int):
                columnRange = [column]
            else:
                columnRange = column

        return columnRange

    def testPulseClose(self, slot):
        # Reset ConfBits
        self.DPXWriteConfigurationCommand(slot, self.confBits[slot-1])

        # Reset peripheryDACs
        self.DPXWritePeripheryDACCommand(slot, self.peripherys + self.THLs[slot-1])

    def maskBitsColumn(self, slot, column):
        # Set ConfBits to use test charge on preamp input if pixel is enabled
        # Only select one column at max
        confBits = np.zeros((16, 16))
        confBits.fill(getattr(ds._ConfBits, 'MaskBit'))
        confBits[column] = [getattr(ds._ConfBits, 'TestBit_Analog')] * 16

        # confBits = np.asarray( [int(num, 16) for num in textwrap.wrap(self.confBits[slot-1], 2)] )
        # confBits[confBits != getattr(ds._ConfBits, 'MaskBit')] = getattr(ds._ConfBits, 'TestBit_Analog')

        self.DPXWriteConfigurationCommand(slot, ''.join(['%02x' % conf for conf in confBits.flatten()]))

    def maskBitsPixel(self, slot, pixel=(0, 0)):
        confBits = np.zeros((16, 16))
        confBits.fill(getattr(ds._ConfBits, 'MaskBit'))
        confBits[pixel[0]][pixel[1]] = getattr(ds._ConfBits, 'TestBit_Analog')
        self.DPXWriteConfigurationCommand(slot, ''.join(['%02x' % conf for conf in confBits.flatten()]))

    def getTestPulseVoltageDAC(self, slot, DACVal, energy=False):
        # Set coarse and fine voltage of test pulse
        peripheryDACcode = int(self.peripherys + self.THLs[slot-1], 16)

        if energy:
            # Use nominal value of test capacitor
            # and DACVal as energy
            C = 5.14e-15

            deltaV = DACVal * scipy.constants.e / (C * 3.62)

            assert deltaV < 1.275, "TestPulse Voltage: The energy of the test pulse was set too high! Has to be less than or equal to 148 keV."

            # Set coarse voltage to 0
            voltageDiv = 2.5e-3
            DACVal = int((1.275 - deltaV) / voltageDiv)
        else:
            assert DACVal >= 0, 'Minimum THL value must be at least 0'
            assert DACVal <= 0x1ff, 'Maximum THL value mustn\'t be greater than %d' % 0x1ff

        # Delete current values
        peripheryDACcode &= ~(0xff << 32)   # coarse
        peripheryDACcode &= ~(0x1ff << 16)  # fine

        # Adjust fine voltage only
        peripheryDACcode |= (DACVal << 16)
        peripheryDACcode |= (0xff << 32)
        # print '%032x' % peripheryDACcode
        # print DACVal

        return '%032x' % peripheryDACcode

    def ToTtoTHL(self, slot=1, column=0, THLstep=1, valueLow=440, valueHigh=512, valueStep=1, energy=False, plot=False, outFn='ToTtoTHL.p'):
        # Description: Generate test pulses and measure their ToT
        #              values. Afterwards, do a THL-scan in order to
        #              find the corresponding THL value. Repeat 
        #              multiple times to find the correlation between
        #              ToT and THL
        # Parameters: 
        #   - energy: if set, iteration is performed over energies.
        #             Else, Test Pulse DAC values are modified directly.

        # Set low level to zero and high level to noise limit
        # for the coarse measurement
        THLlow = 0
        THLhigh = int(self.THLs[slot-1], 16)

        # Set AnalogOut to V_ThA
        OMRCode = self.OMR
        if not isinstance(OMRCode, basestring):
            OMRCode[4] = 'V_ThA'
        else:
            OMRCode = int(OMRCode, 16) & ~(0b11111 << 12)
            OMRCode |= getattr(ds._OMRAnalogOutSel, 'V_ThA')
            OMRCode = '%06x' % OMRCode
        self.DPXWriteOMRCommand(slot, OMRCode)

        valueRange = np.arange(valueLow, valueHigh, valueStep)

        # Number of test pulses for ToT measurement
        NToT = 3

        # Store results per column in dict
        resDict = {}

        # Activate DosiMode and select columns
        columnRange = self.testPulseInit(slot, column=column)

        for column in columnRange:
            THLstart = THLhigh - 1200
            THLstop = THLhigh

            # Select column
            self.maskBitsColumn(slot, column)

            # Store results in lists
            ToTListTotal, ToTErrListTotal = [], []
            THLListTotal, THLErrListTotal = [], []

            # Loop over test pulse energies
            for val in valueRange:
                # Set test pulse energy
                if energy:
                    print('Energy: %.2f keV' % (val/1000.))
                else:
                    val = int(val)
                    print('DAC: %04x' % val)

                # Activates DosiMode and selects columns
                columnRange = self.testPulseInit(slot, column=column)

                # Activate DosiMode
                if not isinstance(OMRCode, basestring):
                    OMRCode[0] = 'DosiMode'
                else:
                    OMRCode = '%06x' % (int(OMRCode, 16) & ~((0b11) << 22))
                self.DPXWriteOMRCommand(slot, OMRCode)
                
                # Set test pulse energy
                DACval = self.getTestPulseVoltageDAC(slot, val, energy)
                self.DPXWritePeripheryDACCommand(slot, DACval)

                # Generate test pulses
                dataList = []
                for i in range(NToT):
                    self.DPXDataResetCommand(slot)
                    self.DPXGeneralTestPulse(slot, 1000)
                    data = self.DPXReadToTDataDosiModeCommand(slot)[column]
                    dataList.append( data )

                # Determine corresponding ToT value
                testPulseToT = np.mean( dataList , axis=0)
                testPulseToTErr = np.std(dataList, axis=0) / np.sqrt( NToT )
                print('Test Pulse ToT:')
                print(testPulseToT)

                # Store in lists
                ToTListTotal.append( testPulseToT )
                ToTErrListTotal.append( testPulseToTErr )

                # Set PCMode
                # WARNING: 8bit counter only!
                if not isinstance(OMRCode, basestring):
                    OMRCode[0] = 'PCMode'
                else:
                    OMRCode = '%06x' % ((int(OMRCode, 16) & ~((0b11) << 22)) | (0b10 << 22))
                self.DPXWriteOMRCommand(slot, OMRCode)
                print(OMRCode)

                # Loop over THLs
                # Init result lists
                lastTHL = 0
                THLList = []
                dataList = []
                THLMeasList = []

                # Dummy readout of AnalogOut
                # self.MCGetADCvalue()
                # THLRange = np.arange(THLlow, THLhigh, THLstep)
                # THLRange = self.THLCalib[np.logical_and(self.THLCalib > THLlow, self.THLCalib < THLhigh)]
                if len(self.THLEdges) == 0 or self.THLEdges[slot - 1] is None:
                    THLRange = np.arange(THLlow, THLhigh)
                else:
                    THLRange = np.asarray(self.THLEdges[slot - 1])
                THLstop_ = THLhigh if THLstop > THLhigh else THLstop
                THLRange = THLRange[np.logical_and(THLRange >= THLstart, THLRange <= THLstop_)]

                # Do a fast measurement to estimate the position of the edge for each pixel
                THLFastList = []
                dataFastList = []
                THLFastStep = 5
                NPulses = 30        # Has to be less than 256

                for cnt, THL in enumerate(THLRange[::THLFastStep]): # :THLFastStep]):
                    # Set new THL
                    self.DPXWritePeripheryDACCommand(slot, DACval[:-4] + '%04x' % THL)

                    # Start measurement
                    self.DPXDataResetCommand(slot)
                    for i in range(NPulses):
                        self.DPXGeneralTestPulse(slot, 1000)
                    data = self.DPXReadToTDatakVpModeCommand(slot)[column]
                    data[data > NPulses] = NPulses

                    # Store data in lists
                    THLFastList.append( THL )
                    dataFastList.append( data )

                if not np.any(THLFastList):
                    continue

                print('THLFastList')
                print(THLFastList)
                print(np.asarray(dataFastList).T)

                # Check if array is empty
                if not np.count_nonzero(dataFastList):
                    # If so, energy is set too low
                    # valueCount -= 1
                    ToTListTotal.pop()
                    ToTErrListTotal.pop()
                    continue

                # Get derivative and find maximum for every pixel
                xFastDer = np.asarray(THLFastList[:-1]) + THLFastStep / 2.
                dataFastDer = [xFastDer[np.argmax(np.diff(data) / float(THLFastStep))] for data in np.asarray(dataFastList).T if np.count_nonzero(data)]

                print('dataFastDer')
                print(dataFastDer)

                # Get mean THL value
                meanTHL = np.mean( dataFastDer )

                # Set THL range to new estimated region
                regSel = np.logical_and(THLRange >= min(dataFastDer) - THLFastStep, THLRange <= max(dataFastDer) + THLFastStep)
                THLRange = THLRange[regSel]
                print('THLRange')
                print(THLRange)
                # THLRange = THLRange[abs(THLRange - meanTHL) < 3 * THLFastStep]

                # Do a slow but precise measurement with small THL steps
                for cnt, THL in enumerate(tqdm(THLRange[::THLstep])):
                    # for V in np.linspace(0.34, 0.36, 1000):
                    # THL = self.getTHLfromVolt(V)
                    # THLmeas = np.mean( [float(int(self.MCGetADCvalue(), 16)) for i in range(3) ] )
                    # THLMeasList.append( THLmeas )
                    # print THL, '%04x' % THL

                    # Set new THL
                    self.DPXWritePeripheryDACCommand(slot, DACval[:-4] + '%04x' % THL)
                    # print DACval[:-4] + '%04x' % THL
                    # print self.peripherys + '%04x' % THL

                    # Generate test pulse and measure ToT in integrationMode
                    # dataTemp = np.zeros(16)
                    self.DPXDataResetCommand(slot)
                    for i in range(NPulses):
                        self.DPXGeneralTestPulse(slot, 1000)
                    data = self.DPXReadToTDatakVpModeCommand(slot)[column]
                    # print data

                    # Store data in lists
                    THLList.append( THL )
                    dataList.append( data )

                    # lastTHL = THLmeas
                print()

                # Calculate derivative of THL spectrum
                xDer = np.asarray(THLList[:-1]) + 0.5*THLstep
                dataDer = [np.diff(data) / float(THLstep) for data in np.asarray(dataList).T]

                peakList = []
                peakErrList = []
                # Transpose to access data of each pixel
                for data in np.asarray(dataList).T:
                    # Convert to corrected THL
                    THLListCorr = [self.getVoltFromTHLFit(elm, slot) for elm in THLList]
                    THLListFit = np.linspace(min(THLListCorr), max(THLListCorr), 1000)

                    # Perform erf-Fit
                    # p0 = [THLListCorr[int(len(THLListCorr) / 2.)], 3.]
                    p0 = [np.mean(THLListCorr), 3.]
                    try:
                        popt, pcov = scipy.optimize.curve_fit(lambda x, b, c: self.erfFit(x, NPulses, b, c), THLListCorr, data, p0=p0)
                        perr = np.sqrt(np.diag(pcov))
                        print(popt)
                    except:
                        popt = p0
                        perr = len(p0) * [0]
                        print('Fit failed!')
                        pass

                    # Return fit parameters
                    # a: Amplitude
                    # b: x-offset
                    # c: scale
                    b, c = popt
                    peakList.append( b )
                    # Convert to sigma
                    peakErrList.append( perr[1] / np.sqrt(2) )

                    # Savitzky-Golay filter the data
                    # Ensure odd window length
                    windowLength = int(len(THLRange[::THLstep]) / 10.)
                    if not windowLength % 2:
                        windowLength += 1

                    # Plots
                    if plot:
                        try:
                            dataFilt = scipy.signal.savgol_filter(data, windowLength, 3)
                            plt.plot(THLListCorr, dataFilt)
                            plt.plot(*self.getDerivative(THLListCorr, dataFilt))
                        except:
                            pass

                        plt.plot(THLListFit, self.erfFit(THLListFit, NPulses, b, c), ls='-')
                        plt.plot(THLListFit, self.normalErf(THLListFit, NPulses, b, c), ls='-')
                        plt.plot(THLListCorr, data, marker='x', ls='')
                        plt.show()

                peakList, peakErrList = np.asarray(peakList), np.asarray(peakErrList)
                peakList[peakErrList > 50] = np.nan
                peakErrList[peakErrList > 50] = np.nan
                print(peakList, peakErrList)

                # Set start value for next scan
                print('Start, Stop')
                THLstart, THLstop = np.nanmin(THLList) - 400, np.nanmax(THLList) + 400
                print(THLstart, THLstop)

                # Convert back to THL
                # peakList = [THLRange[int(peak + 0.5)] + (peak % 1) for peak in peakList]

                THLListTotal.append( peakList )
                THLErrListTotal.append( peakErrList )
                print()

            # Transform to arrays
            THLListTotal = np.asarray(THLListTotal).T
            THLErrListTotal = np.asarray(THLErrListTotal).T
            ToTListTotal = np.asarray(ToTListTotal).T
            ToTErrListTotal = np.asarray(ToTErrListTotal).T

            print(THLListTotal)
            print(THLErrListTotal)
            print(ToTListTotal)
            print(ToTErrListTotal)

            resDict[column] = {'THL': THLListTotal, 'THLErr': THLErrListTotal, 'ToT': ToTListTotal, 'ToTErr': ToTErrListTotal, 'volt': valueRange}

        if plot:
            # Show results in plot for last accessed column
            fig, ax = plt.subplots()
            for i in range(len(THLListTotal)):
                ax.errorbar(THLListTotal[i], ToTListTotal[i], xerr=THLErrListTotal[i], yerr=ToTErrListTotal[i], color=self.getColor('Blues', len(ToTListTotal), i), marker='x')

            ax.grid()
            plt.xlabel('THL (DAC)')
            plt.ylabel('ToT')

            # TODO: Perform fit
            plt.show()

        # Save to pickle file
        if outFn:
            # Create dictionary
            # d = {'THL': THLListTotal, 'THLErr': THLErrListTotal, 'ToT': ToTListTotal, 'ToTErr': ToTErrListTotal}

            self.pickleDump(resDict, outFn)

        return resDict

    def ToTtoTHL_pixelDAC(self, slot=1, THLstep=1, I_pixeldac=0.1, valueLow=0, valueHigh=460, valueCount=460, energy=False, plot=False):
        pixelDACs = ['00', '3f']

        # Set I_pixeldac
        if I_pixeldac:
            dPeripherys = self.splitPerihperyDACs(self.peripherys + self.THLs[0], perc=True)
            dPeripherys['I_pixeldac'] = I_pixeldac
            code = self.periheryDACsDictToCode(dPeripherys, perc=True)
            self.peripherys = code[:-4]
            self.DPXWritePeripheryDACCommand(slot, code)

        dMeas = {}
        slopeDict = {'slope': []}

        for column in range(16):
            for pixelDAC in pixelDACs:
                # Set pixel DAC values
                self.DPXWritePixelDACCommand(slot, pixelDAC*256, file=False)

                d = self.ToTtoTHL(slot=slot, column=column, THLstep=THLstep, valueLow=valueLow, valueHigh=valueHigh, valueCount=valueCount, energy=energy, plot=plot)

                # Get THL positions
                dMeas[pixelDAC] = d['THL']

            # Get slope
            slope = (dMeas['00'] - dMeas['3f']) / 63.

            print('Pixel slope for column %d:' % column)
            print(slope)
            print('Mean slope: %.2f +/- %.2f' % (np.mean(slope), np.std(slope)))

            slopeDict['slope'].append( slope )

        self.pickleDump(slopeDict, 'pixelSlopes.p')

        return slope

    def TPtoToTPixel(self, slot=1, column=0, low=400, high=512, step=5, outFn='TPtoToT.p'):
        # Description: Generate test pulses and measure ToT afterwards
        #              to match test pulse voltage and ToT value.

        # Number of test pulses for ToT measurement
        NToT = 3

        # Test pulse voltage range
        # TPvoltageRange = list(reversed(np.arange(300, 490, 1))) # list(reversed(np.arange(490, 512, 1)))
        # TPvoltageRange = list(reversed(np.arange(470, 505, 2))) + list(reversed(np.arange(350, 470, 10)))
        TPvoltageRange = list(reversed(np.arange(low, high, step))) 

        # Store results per column in dict
        resDict = {}

        # Activate DosiMode and select columns
        columnRange = self.testPulseInit(slot, column=column)

        import time
        # Deselect all pixels
        confBits = np.zeros((16, 16))
        confBits.fill(getattr(ds._ConfBits, 'MaskBit'))

        ToTListTotal, ToTErrListTotal = [], []
        for val in TPvoltageRange:
            start_time = time.time()

            # Set test pulse voltage
            DACval = self.getTestPulseVoltageDAC(slot, val, energy=False)
            self.DPXWritePeripheryDACCommand(slot, DACval)

            dataList = []
            lastpixelx, lastpixely = 15, 15
            for i in range(NToT):
                self.DPXDataResetCommand(slot)
                for pixelx in range(16):
                    for pixely in range(16):
                        confBits[lastpixelx, lastpixely] = getattr(ds._ConfBits, 'MaskBit')
                        confBits[pixelx, pixely] = getattr(ds._ConfBits, 'TestBit_Analog')
                        self.DPXWriteConfigurationCommand(slot, ''.join(['%02x' % conf for conf in confBits.flatten()]))
                        self.DPXGeneralTestPulse(slot, 1000)
                        lastpixelx, lastpixely = pixelx, pixely
                data = np.asarray(self.DPXReadToTDataDosiModeCommand(slot), dtype=float)
                data[data > 1000] = np.nan
                dataList.append( data )

            ToT, ToTErr = np.nanmedian(dataList, axis=0), np.nanstd(dataList, axis=0) / np.sqrt(NToT)
            print(ToT[0])
            ToTListTotal.append( ToT ), ToTErrListTotal.append( ToTErr )
            print('Time:', (time.time() - start_time))
        resDict = [{'ToT': np.asarray(ToTListTotal)[:,column,:], 'ToTErr': np.asarray(ToTErrListTotal)[:,column,:], 'volt': TPvoltageRange} for column in range(16)]

        outFn = outFn.split('.')[0] + '.json' % T
        self.pickleDump(resDict, outFn)

    def TPTime(self, slot=1, column=0, voltage=100, timeRange=np.logspace(1, 3, 10), outFn='TPTime.json'):
        NToT = 10
        columnRange = self.testPulseInit(slot, column=column)
        DACval = self.getTestPulseVoltageDAC(slot, voltage, energy=False)
        self.DPXWritePeripheryDACCommand(slot, DACval)

        resDict = {column: {'ToT': [], 'ToTErr': [], 'time': timeRange} for column in columnRange}
        confBits = np.zeros((16, 16))
        for time in timeRange:
            print(time)
            for column in columnRange:
                confBits.fill(getattr(ds._ConfBits, 'MaskBit'))
                confBits[column] = [getattr(ds._ConfBits, 'TestBit_Analog')] * 16 + [getattr(ds._ConfBits, 'MaskBit')] * 0
                self.DPXWriteConfigurationCommand(slot, ''.join(['%02x' % conf for conf in confBits.flatten()]))

                dataList = []
                for i in range(NToT):
                    self.DPXDataResetCommand(slot)
                    self.DPXGeneralTestPulse(slot, time)
                    data = self.DPXReadToTDataDosiModeCommand(slot)
                    dataList.append( data[column] )

                ToT, ToTErr = np.median(dataList, axis=0), np.std(dataList, axis=0) / np.sqrt(NToT)

            resDict[column]['ToT'].append(ToT)
            resDict[column]['ToTErr'].append(ToTErr)

        self.pickleDump(resDict, outFn)

    def TPfindMax(self, slot=1, column=0):
        voltRange = np.arange(100, 400, 50)
        NToT = 3
        columnRange = self.testPulseInit(slot, column=column)
        confBits = np.zeros((16, 16))

        xMaxList = []
        for idx, volt in enumerate( voltRange ): 
            print('Scanning Test Pulse energy %d...' % volt)
            # Set test pulse voltage
            DACval = self.getTestPulseVoltageDAC(slot, volt, energy=False)
            self.DPXWritePeripheryDACCommand(slot, DACval)

            time = 1.
            x, y = [], []
            max_val = 0
            low_cnt = 0
            step_size = 5
            start_flag = True
            scans = 0
            it, maxIterations = 0, 30
            while True:
                columnList = []
                for column in columnRange:
                    confBits.fill(getattr(ds._ConfBits, 'MaskBit'))
                    confBits[column] = [getattr(ds._ConfBits, 'TestBit_Analog')] * 16 + [getattr(ds._ConfBits, 'MaskBit')] * 0
                    self.DPXWriteConfigurationCommand(slot, ''.join(['%02x' % conf for conf in confBits.flatten()]))

                    dataList = []
                    for i in range(NToT):
                        self.DPXDataResetCommand(slot)
                        self.DPXGeneralTestPulse(slot, time)
                        data = self.DPXReadToTDataDosiModeCommand(slot)
                        dataList.append( data[column] )

                    ToT, ToTErr = np.median(dataList, axis=0), np.std(dataList, axis=0) / np.sqrt(NToT)
                    columnList.append( ToT )
                
                x.append( time ), y.append( np.median(columnList) )
                it += 1
                if start_flag:
                    # Found peak
                    if len(y) > 2 and (y[-3] < y[-2]) and (y[-2] > y[-1]):
                        start_flag = False
                        time -= 1.5 * step_size
                        x = x[-3:]
                        y = y[-3:]
                        continue
                    else:
                        time += step_size
                else:
                    scans += 1
                    if scans == 2:
                        scans = 0

                        sort_idx = np.argsort(x)
                        x, y = list(np.asarray(x)[sort_idx][-5:]), list(np.asarray(y)[sort_idx][-5:])
                        if np.std(y) < 1.5 or (it > maxIterations):
                            break
                
                        zero = np.where(np.diff(np.sign(np.diff(y))))[0][0]

                        x = x[zero:zero+3]
                        y = y[zero:zero+3]

                        step_size *= 0.5
                        time = x[0] + 0.5 * step_size

                    else:
                        time += step_size

                continue

            # plt.plot(x, y, marker='x', ls='')
            spl = scipy.interpolate.UnivariateSpline(x, y, s=len(x) * 10, k=4)
            roots = spl.derivative().roots()
            x_max = roots[np.argmax(spl(roots))]

            # plt.plot(x, spl(x))
            # plt.plot(x_max, spl(x_max), marker='x', ls='', markersize=10, label=volt)

            xMaxList.append(x_max)
            # plt.show()

        print('Scanned Test Pulse energies:', voltRange)
        print('Maxima at:', xMaxList)

        idx = np.asarray(xMaxList[-2:]) > np.median(xMaxList[:2])
        idx = np.asarray([True] * (len(xMaxList) - 2) + list(idx))

        popt, pcov = scipy.optimize.curve_fit(self.linearFit, np.asarray(voltRange)[idx], np.asarray(xMaxList)[idx])
        perr = np.sqrt(np.diag(pcov))
        print('Test Pulse energy to Test Pulse duration relation:')
        print('t = (%.2f +/- %.2f) * E_TP + (%.2f +/- %.2f)' % (popt[0], perr[0], popt[1], perr[1]))
        return popt

    def TPtoToT(self, slot=1, column=0, low=400, high=512, step=5, outFn='TPtoToT.p'):
        # Description: Generate test pulses and measure ToT afterwards
        #              to match test pulse voltage and ToT value.

        # Number of test pulses for ToT measurement
        NToT = 10

        TPvoltageRange = list(reversed(np.arange(low, high, step))) 

        # Activate DosiMode and select columns
        columnRange = self.testPulseInit(slot, column=column)

        # Store results per column in dict
        resDict = {column: {'ToT': [], 'ToTErr': [], 'volt': TPvoltageRange} for column in columnRange}

        # Get times for maximum ToT
        print('Optimizing test pulse durations...')
        pTime = self.TPfindMax(slot=slot, column=0)

        print('Done!')
        print()

        import time
        it_idx = 0
        print('Test Pulse to ToT measurement:')
        for val in TPvoltageRange:
            start_time = time.time()

            # Set test pulse voltage
            DACval = self.getTestPulseVoltageDAC(slot, val, energy=False)
            self.DPXWritePeripheryDACCommand(slot, DACval)

            # Deselect all pixels
            confBits = np.zeros((16, 16))

            for column in tqdm(columnRange, desc='Column'):
                confBits.fill(getattr(ds._ConfBits, 'MaskBit'))
                confBits[column] = [getattr(ds._ConfBits, 'TestBit_Analog')] * 16 + [getattr(ds._ConfBits, 'MaskBit')] * 0
                self.DPXWriteConfigurationCommand(slot, ''.join(['%02x' % conf for conf in confBits.flatten()]))

                dataList = []
                for i in tqdm(range(NToT), leave=False):
                    self.DPXDataResetCommand(slot)
                    m, t = pTime
                    self.DPXGeneralTestPulse(slot, m * val + t)
                    data = self.DPXReadToTDataDosiModeCommand(slot)
                    dataList.append( data[column] )

                    it_idx += 1
                # print np.median(dataList, axis=0)

                ToT, ToTErr = np.median(dataList, axis=0), np.std(dataList, axis=0) / np.sqrt(NToT)

                resDict[column]['ToT'].append(ToT)
                resDict[column]['ToTErr'].append(ToTErr)

        self.pickleDump(resDict, outFn)

    def testPulseToT(self, slot, length, column='all', paramOutFn='testPulseParams.p', DAC=None, DACRange=None, perc=True):
        columnRange = self.testPulseInit(slot, column=column)

        # Measure multiple test pulses and return the average ToT
        energyRange = np.asarray(list(np.linspace(1.e3, 20.e3, 30)) + list(np.linspace(20e3, 100e3, 20)))
        energyRangeFit = np.linspace(1.e3, 100e3, 1000)
        # energyRange = np.arange(509, 0, -10)

        outDict = {'energy': energyRange, 'DAC': DACRange}

        # Set DAC value
        if DAC is not None:
            if DACRange is not None:
                outList = []
                for DACVal in DACRange:
                    d = self.splitPerihperyDACs(self.peripherys + self.THLs[0], perc=perc)
                    d[DAC] = DACVal
                    code = self.periheryDACsDictToCode(d, perc=perc)
                    self.peripherys = code[:-4]
                    self.DPXWritePeripheryDACCommand(slot, code)

                    dataArray = []
                    for column in  columnRange:
                        self.maskBitsColumn(slot, column)

                        ToTValues = []
                        for energy in energyRange:
                            # DACval = self.getTestPulseVoltageDAC(slot, energy)
                            DACval = self.getTestPulseVoltageDAC(slot, energy, True)
                            self.DPXWritePeripheryDACCommand(slot, DACval)

                            data = []
                            for i in range(10):
                                self.DPXDataResetCommand(slot)
                                self.DPXGeneralTestPulse(slot, 1000)
                                data.append( self.DPXReadToTDataDosiModeCommand(slot)[column] )
                            data = np.mean(data, axis=0)

                            # Filter zero entries
                            # data = data[data != 0]

                            ToTValues.append( data )

                        ToTValues = np.asarray( ToTValues )
                        print(ToTValues)
                        dataArray.append( ToTValues.T )
                        print(dataArray)

                        if False:
                            x = energyRange/float(1000)
                            for row in range(16):
                                y = []
                                for ToTValue in ToTValues:
                                    # print ToTValue
                                    y.append( ToTValue[row])

                                # Fit curve
                                try:
                                    popt, pcov = scipy.optimize.curve_fit(self.energyToToTFitAtan, x, y, p0=(3, 8, 50, 1))
                                    perr = np.sqrt( np.diag(pcov) )
                                except:
                                    popt = (3, 8, 50, 1)
                                    perr = np.zeros(len(popt))

                                # Store in dictionary
                                a, b, c, t = popt
                                paramOutDict['a'].append( a )
                                paramOutDict['b'].append( b )
                                paramOutDict['c'].append( c )
                                paramOutDict['t'].append( t )

                                print(popt, perr/popt*100)

                                plt.plot(x, y, marker='x')
                                plt.plot(energyRangeFit/float(1000), self.energyToToTFitAtan(energyRangeFit/float(1000), *popt))
                                plt.show()

                    outList.append(np.vstack(dataArray))

        outDict['data'] = outList

        # Dump to file
        if paramOutFn:
            json.dump(outDict, open(paramOutFn, 'wb'))
        return

        # Plot
        plt.xlabel('Energy (keV)')
        plt.ylabel('ToT')
        plt.grid()
        plt.show()

        self.testPulseClose(slot)
        return

    def measureTestPulses(self, slot, column='all'):
        columnRange = self.testPulseInit(slot, column=column)
        for column in columnRange:
            self.maskBitsColumn(slot, column)

            while True:
                self.DPXDataResetCommand(slot)
                # self.DPXGeneralTestPulse(slot, 1000)
                print(self.DPXReadToTDataDosiModeCommand(slot)[column])

    def measurePulseShape(self, slot, column='all'):
        assert len(self.THLEdges) > 0, 'Need THL calibration first!'

        columnRange = self.testPulseInit(slot, column=column)
        N = 10

        # Set energy
        energy = 10.e3
        DACVal = self.getTestPulseVoltageDAC(slot, energy, True)
        print('Energy DAC:', DACVal)
        # self.DPXWritePeripheryDACCommand(slot, DACval)

        # Define THLRange
        THLhigh = int(self.THLs[slot-1], 16)
        THLstep = 1
        THLstart = THLhigh - 1000
        THLstop = THLhigh

        THLRange = np.asarray(self.THLEdges[slot - 1])
        THLstop_ = THLhigh if THLstop > THLhigh else THLstop
        THLRange = THLRange[np.logical_and(THLRange >= THLstart, THLRange <= THLstop_)]

        for column in columnRange:
            self.maskBitsColumn(slot, column)

            ToTList, voltList = [], []
            for cnt, THL in enumerate(tqdm(THLRange[::THLstep])):

                # Set new THL
                self.DPXWritePeripheryDACCommand(slot, DACVal[:-4] + '%04x' % THL)

                dataList = []
                for n in range(N):
                    self.DPXDataResetCommand(slot)
                    self.DPXGeneralTestPulse(slot, 1000)
                    dataList.append( self.DPXReadToTDataDosiModeCommand(slot)[column] )

                mean = np.mean( dataList, axis=0 )
                if np.any(mean):
                    print(mean)
                    ToTList.append( mean )
                    voltList.append( self.getVoltFromTHLFit(THL, slot) )

            plt.plot(ToTList, -np.asarray(voltList))
            plt.show()

        self.testPulseClose(slot)

    def testPulseSigma(self, slot, column=[0, 1, 2], N=100):
        columnRange = self.testPulseInit(slot, column=column)

        fig, ax = plt.subplots()

        energyRange = np.linspace(5e3, 100e3, 50)
        meanListEnergy, sigmaListEnergy = [], []

        for energy in energyRange:
            DACval = self.getTestPulseVoltageDAC(slot, energy, True)
            print('Energy DAC:', DACVal)
            self.DPXWritePeripheryDACCommand(slot, DACval)

            meanList, sigmaList = [], []
            print('E = %.2f keV' % (energy / 1000.))
            for k, column in enumerate(tqdm(columnRange), desc='Column'):
                self.maskBitsColumn(slot, column)

                # Record ToT spectrum of pulse
                dataList = []
                for n in range(tqdm(N), leave=False):
                    self.DPXDataResetCommand(slot)
                    self.DPXGeneralTestPulse(slot, 100)
                    dataList.append( self.DPXReadToTDataDosiModeCommand(slot)[column] )

                dataList = np.asarray(dataList)
                for pixel in range(16):
                    # ax.hist(dataList[:,pixel], bins=30)
                    meanList.append( np.mean(dataList[:,pixel]) )
                    sigmaList.append( np.std(dataList[:,pixel]) )
                    # plt.show()

            print()
            print(meanList)
            meanListEnergy.append( meanList )
            sigmaListEnergy.append( sigmaList )

        meanListEnergy = np.asarray(meanListEnergy)
        sigmaListEnergy = np.asarray(sigmaListEnergy)
        # for pixel in range(len(meanListEnergy[0])):

        plt.errorbar(energyRange/1000., [np.mean(item) for item in sigmaListEnergy], xerr=energyRange/1000.*0.18, yerr=np.asarray([np.std(item) for item in sigmaListEnergy])/np.sqrt(len(columnRange)*16), marker='x', ls='')

        # Fit curve
        #popt, pcov = scipy.optimize.curve_fit(self.energyToToTFitAtan, energyRange/1000., [np.mean(item) for item in meanListEnergy], p0=(3, 8, 50, 1))
        #perr = np.sqrt( np.diag(pcov) )
        #print popt, perr/popt*100

        #plt.plot(energyRange/float(1000), self.energyToToTFitAtan(energyRange/float(1000), *popt))

        plt.xlabel('Energy (keV)')
        plt.ylabel('Standard deviation (ToT)')
        plt.show()

        self.testPulseClose(slot)

    def testPulseDosi(self, slot=1, column='all'):
        columnRange = self.testPulseInit(slot, column=column)

        # Select DosiMode
        '''
        OMRCode = self.OMR
        if not isinstance(OMRCode, basestring):
            OMRCode[0] = 'DosiMode'
        else:
            OMRCode = '%04x' % (int(OMRCode, 16) & ~((0b11) << 22))

        self.DPXWriteOMRCommand(slot, OMRCode)
        '''

        # Specify energy range
        energyRange = [25.e3] # list(np.linspace(20e3, 100e3, 10))

        ToTValues = []
        # Loop over energies
        for energy in energyRange:
            DACval = self.getTestPulseVoltageDAC(slot, energy, True)
            self.DPXWritePeripheryDACCommand(slot, DACval)

            # Clear data
            self.clearBins([slot])

            # Loop over columns
            outList = []
            for column in columnRange:
                self.maskBitsColumn(slot, 15 - column)
                # self.DPXWriteColSelCommand(slot, column)

                # Dummy readouts
                # self.DPXReadBinDataDosiModeCommand(slot)

                # Generate pulses
                for i in range(10):
                    # time.sleep(1)
                    self.DPXGeneralTestPulse(slot, 1000)

                # Read from column
                self.DPXWriteColSelCommand(slot, column)
                out = self.DPXReadBinDataDosiModeCommand(slot)
                print(np.mean(out), np.std(out))
                print(out)
                outList.append( out )

            dataMatrix = np.rec.fromarrays( outList )
            for i in range(len(dataMatrix)):
                print(np.asarray([list(entry) for entry in dataMatrix[i]]))
                plt.imshow(np.asarray([list(entry) for entry in dataMatrix[i]]))
                plt.show()

        self.testPulseClose(slot)
        return




#!/usr/bin/env python
import dpx_control
import json

# Important files
THL_CALIB_FILES = ['THLCalibration/THLCalib_%d.p' % slot for slot in [22, 6, 109]]
BIN_EDGES_FILE = None # 'Dennis1_binEdges.hck'
PARAMS_FILE = 'energyConversion/paramsDict_DPX22_6_109.hck'

GEN_BIN_EDGES = False
GEN_BIN_EDGES_RANDOM = False
GEN_BIN_EDGES_UNIFORM = True
# BIN_EDGES_RANDOM_FN = 'binEdgesRandom_DPX22_6_109_v2.hck'
BIN_EDGES_RANDOM_FN = 'binEdgesUniform_DPX22_6_109_v2.hck'

if GEN_BIN_EDGES:
    import bin_edges_random as ber

    if PARAMS_FILE.endswith('.p'):
        paramsDict = json.load(open(PARAMS_FILE, 'rb'))

    binEdgesDict = {}
    for slot in range(1, 3 + 1):
        if GEN_BIN_EDGES_RANDOM:
            binEdges = ber.getBinEdgesRandom(NPixels=256, edgeMin=12, edgeMax=100, edgeOvfw=430, uniform=False)
        elif GEN_BIN_EDGES_UNIFORM:
	    # Generate edges for multiple energy regions
            binEdges = []
	    energy_start, energy_range = 10, 90
	    for idx in range(4):
	    	edges = ber.getBinEdgesUniform(NPixels=256, edgeMin=energy_start + idx*energy_range, edgeMax=energy_start + (idx + 1)*energy_range, edgeOvfw=430)
		binEdges.append( edges )
        binEdgesDict['Slot%d' % slot] = binEdges
    json.dump(binEdgesDict, BIN_EDGES_RANDOM_FN)

BIN_EDGES = {'Slot1': [12, 18, 21, 24.5, 33.5, 43, 53.5, 66.5, 81.5, 97, 113, 131.5, 151.5, 173, 200.5, 236, 430],
                'Slot2': [12, 17, 31, 40, 45.5, 50.5, 60.5, 68, 91, 102.5, 133, 148, 163, 196, 220, 257, 430],
                'Slot3': [32, 37, 47, 57.6, 68.5, 80, 91.5, 104, 117, 131, 145, 163.5, 183.5, 207.5, 234.5, 269.5, 430]}
PORT = '/dev/tty.usbserial-A907PD5F'
CONFIG_FN = 'DPXConfig.conf'

def main():
    dpx = dpx_control.Dosepix(PORT, 2e6, CONFIG_FN, bin_edges_file=BIN_EDGES, params_file=PARAMS_FILE, thl_calib_files=THL_CALIB_FILES)
    dpx.measureDose(slot=1, measurement_time=0, freq=False, frames=1000, logTemp=False, intPlot=True, conversion_factors=None)
    dpx.close()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
import dpx_control

PORT = '/dev/ttyUSB0'
CONFIG_DIR = 'config/'
CHIP_NUMS = [22, 6, 109]
IKRUM = 20
CONFIG_FN = [CONFIG_DIR + '%d/DPXConfig_%d_Ikrum%d.conf' % (CHIP, CHIP, IKRUM) if CHIP is not None else None for CHIP in CHIP_NUMS]
PARAMS_FILES = ['calibration_parameters/params_%d_Ikrum%d.json' % (CHIP, IKRUM) if CHIP is not None else None for CHIP in CHIP_NUMS]
BIN_EDGES_FILES = ['bin_edges/bin_edges_uniform_energy_10_80_4splits.json'] * 3

def main():
    # Establish connection
    thl_calib_files = None 
    dpx = dpx_control.Dosepix(PORT, 2e6, CONFIG_FN, thl_calib_files=thl_calib_files,
            params_file=PARAMS_FILES, bin_edges_file=BIN_EDGES_FILES)
    # dpx.measureADC(slot=1, AnalogOut='I_krum', perc=False, ADChigh=4096, ADClow=0, ADCstep=1, N=1, fn='Ikrum_meas.json', plot=True)
    dpx.ADCWatch(slot=1, OMRAnalogOutList=['I_krum'], cnt=0, fn='ADCWatch.json')

    # Close connection
    dpx.close()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
import dpx_control
import numpy as np

REUSE_CONFIG = False
PORT = '/dev/ttyUSB0'
CONFIG_DIR = 'config/'
CHIP_NUMS = [22, None, None] # 101, 109]

if REUSE_CONFIG:
    IKRUMS = np.arange(37, 51) 
else:
    IKRUMS = [20, 40, 50] # [10, 20, 30, 40, 50]

PARAMS_FILES = None
BIN_EDGES_FILES = None
MEAS_TIME = 60 * 30

def main():
    if REUSE_CONFIG:
        CONFIG_FN = [CONFIG_DIR + '%d/DPXConfig_%d_Ikrum%d.conf' % (CHIP, CHIP, 20) if CHIP is not None else None for CHIP in CHIP_NUMS]

    # Loop over different Ikrum values and perform measurements
    for Ikrum in IKRUMS:
        if not REUSE_CONFIG:
            CONFIG_FN = [CONFIG_DIR + '%d/DPXConfig_%d_Ikrum%d.conf' % (CHIP, CHIP, Ikrum) if CHIP is not None else None for CHIP in CHIP_NUMS]

        print('=== Starting ToT Measurement for Ikrum %d ===' % Ikrum)
        # Establish connection
        thl_calib_files = None 
        dpx = dpx_control.Dosepix(PORT, 2e6, CONFIG_FN, thl_calib_files=thl_calib_files, 
                params_file=PARAMS_FILES, bin_edges_file=BIN_EDGES_FILES)
        
        # Change Ikrum
        if REUSE_CONFIG:
            # Ikrum is changed in this section, else it is changed via configuration file
            for chip_idx in range(3):
                d = dpx.splitPerihperyDACs(dpx.peripherys[chip_idx] + dpx.THLs[chip_idx], perc=False, show=True)
                d['I_krum'] = Ikrum
                code = dpx.periheryDACsDictToCode(d, perc=False)
                code = code[:-4] + '%04x' % (int(dpx.THLs[chip_idx], 16))

                dpx.peripherys[chip_idx] = code[:-4]
                dpx.DPXWritePeripheryDACCommand(chip_idx + 1, code)

        # Measure ToT
        dpx.measureToT(slot=[1, 2, 3], intPlot=False, cnt=10000, storeEmpty=False, logTemp=True, meas_time=MEAS_TIME,
                        outDir='ToTMeasurement_%s_Ikrum%d/' % ('_'.join([str(chip) for chip in CHIP_NUMS]), Ikrum))

        # Close connection
        dpx.close()
        print()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
import dpx_control
import dpx_control.multi as mu

PORT = ['/dev/ttyUSB0']
CONFIG_DIR = ['config/']
CHIP_NUMS = [[22, 6, 109]]
# PORT = ['/dev/ttyUSB0', '/dev/ttyUSB1']
# CONFIG_DIR = ['config/', 'config/']
# CHIP_NUMS = [[22, 6, 109], [101, 154, None]]
CONFIG_FN = [[CONFIG_DIR[idx] + 'DPXConfig_%d.conf' % CHIP if CHIP is not None else None for CHIP in CHIP_NUMS[idx]] for idx in range(len(CONFIG_DIR))]
N_DET = len(PORT)
thl_calib_files = None 

def main():
    # Establish connection
    dpxObjects = []
    for idx in range(N_DET): 
        print(CONFIG_FN[idx])
        dpxObjects.append( dpx_control.Dosepix(PORT[idx], 2e6, CONFIG_FN[idx], thl_calib_files=thl_calib_files) )

    dpx_multi = mu.DosepixMulti(dpxObjects, [[1]])
    dpx_multi.measureDose(frames=10000, sync=True)

    # Close connection
    for dpx in dpxObjects:
        dpx.close()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
import numpy as np

def getBinEdgesRandom(NPixels, edgeMin, edgeMax, edgeOvfw, uniform=False):
    edgeList = []
    for pixel in range(NPixels):
        if uniform:
            # Calculate bin edges
            binEdges = np.sort(np.random.uniform(edgeMin, edgeMax, 15))
            binEdges = np.insert(binEdges, 0, edgeMin)
            binEdges = np.append(binEdges, edgeOvfw)
        else:
            # Get bin edges with noisy evenly spaced distances
            binEdges = getBinEdgesRandomEvenSpace(edgeMin, edgeMax, edgeOvfw)

            if any(np.isnan(binEdges)):
                binEdges = getBinEdgesRandomEvenSpace(edgeMin, edgeMax, edgeOvfw)

        # Convert to ToT
        edgeList.append( binEdges )
    return edgeList

def getBinEdgesRandomEvenSpace(edgeMin, edgeMax, edgeOvfw):
    # Mean difference 
    diff = float(edgeMax - edgeMin) / 15
    binDiff = np.random.normal(diff, diff / 3., 15)
    binDiff -= (np.sum(binDiff) - (edgeMax - edgeMin)) / 15
    binEdges = np.cumsum(binDiff) + edgeMin
    binEdges = np.insert(binEdges, 0, 1.5 * edgeMin + abs(np.random.normal(0, 0.5*diff)))
    binEdges = np.append(binEdges, edgeOvfw)
    binEdges = np.sort( binEdges )

    return binEdges

def getBinEdgesUniform(pixels, edgeMin, edgeMax, edgeOvfw):
    edgeList = []
    xInit = np.linspace(edgeMin, edgeMax, 16)
    bw = xInit[1] - xInit[0]
    pixelIdx = 0

    pixels_filt = [pixel for pixel in pixels if isLarge(pixel)]
    for pixel in pixels:
        if not isLarge(pixel):
            edgeList.append(np.append(xInit, edgeOvfw))
            continue

        offset = bw / float(len(pixels_filt)) * pixelIdx
        edgeList.append(np.append(xInit + offset, edgeOvfw))
        pixelIdx += 1
    return edgeList

def isLarge(pixel):
	if pixel % 16 in [0, 1, 14, 15]:
		return False
	else:
		return True

# Conversion functions
# Energy to ToT and vice versa
def ToTtoEnergySimple(self, x, a, b, c, t, h=1, k=0):
    return h * (b + 1./(4 * a) * (2*x + np.pi*c + np.sqrt(16 * a * c * t + (2 * x + np.pi * c)**2))) + k

def EnergyToToTSimple(x, a, b, c, t, h=1, k=0):
    res = np.where(x < b, a*((x - k)/h - b) - c * (np.pi / 2 + t / ((x - k)/h - b)), 0)
    res[res < 0] = 0
    return res



from __future__ import print_function
import json
import numpy as np
import textwrap
import serial
import os
import sys
import matplotlib.pyplot as plt
import matplotlib as mpl
import scipy.special

class Support(object):
    def normal(self, x, A, mu, sigma):
        return A * np.exp(-(x - mu)/(2 * sigma**2))

    def getColor(self, c, N, idx):
        import matplotlib as mpl
        cmap = mpl.cm.get_cmap(c)
        norm = mpl.colors.Normalize(vmin=0.0, vmax=N - 1)
        return cmap(norm(idx))

    def getColorBar(self, ax, cbMin, cbMax, N=20, label=None, rotation=90):
        # Plot colorbar
        from matplotlib.colors import ListedColormap
        cmap = mpl.cm.get_cmap('viridis', N)
        norm = mpl.colors.Normalize(vmin=cbMin, vmax=cbMax)
        cBar = mpl.colorbar.ColorbarBase(ax, cmap=cmap, norm=norm, orientation='vertical')

        # cBar.ax.invert_yaxis()
        cBar.formatter.set_powerlimits((0, 0))
        cBar.ax.yaxis.set_offset_position('right')
        cBar.update_ticks()

        labels = np.linspace(cbMin, cbMax, N + 1)
        locLabels = np.linspace(cbMin, cbMax, N)
        loc = labels + abs(labels[1] - labels[0])*.5
        cBar.set_ticks(loc)
        cBar.ax.set_yticklabels(['%.1f' % loc for loc in locLabels], rotation=rotation, verticalalignment='center')
        cBar.outline.set_visible(False)
        cBar.set_label(label)

    # https://stackoverflow.com/questions/12090503/listing-available-com-ports-with-python
    def getSerialPorts(self):
        """ Lists serial port names

            :raises EnvironmentError:
                On unsupported or unknown platforms
            :returns:
                A list of the serial ports available on the system
        """
        if sys.platform.startswith('win'):
            ports = ['COM%s' % (i + 1) for i in range(256)]
        elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
            # this excludes your current terminal "/dev/tty"
            ports = glob.glob('/dev/tty[A-Za-z]*')
        elif sys.platform.startswith('darwin'):
            ports = glob.glob('/dev/tty.*')
        else:
            raise EnvironmentError('Unsupported platform')

        result = []
        for port in ports:
            try:
                s = serial.Serial(port)
                s.close()
                result.append(port)
            except (OSError, serial.SerialException):
                pass
        return result

    def getDerivative(self, x, y):
        deltaX = x[1] - x[0]
        der = np.diff(y) / deltaX
        return x[:-1] + deltaX, der

    def erfFit(self, x, a, b, c):
        return a*(0.5 * scipy.special.erf((x - b)/c) + 0.5)

    def normalErf(self, x, a, b, c):
        return 0.56419*a*np.exp(-(x-b)**2/(c**2)) + 0.5

    def linearFit(self, x, m, t):
        return m*x + t

    def erfStdFit(self, x, a, b, c, d):
        return a * (scipy.special.erf((x - b)/c) + 1) + d

    def meanSigmaLinePlot(self, fn, pIpixeldac, pPixelDAC):
        d = json.load( open(fn, 'rb') )
        meanMatrix, sigmaMatrix = d['mean'], d['sigma']
        meanMatrix, sigmaMatrix = meanMatrix, sigmaMatrix

        # Ipixeldac params
        minIpixeldac, maxIpixeldac, NIpixeldac = pIpixeldac
        # pixelDAC params
        minDAC, maxDAC, NDAC = pPixelDAC

        # Create x-axis
        pixelDAC = np.linspace(minDAC, maxDAC, NDAC)

        figMean = plt.figure(figsize=(7, 7))
        axMean = figMean.add_axes([0.1, 0.14, 0.7, 0.8])
        axCBarMean = figMean.add_axes([0.85, 0.1, 0.05, 0.8])
        figSigma = plt.figure(figsize=(7, 7))
        axSigma = figSigma.add_axes([0.1, 0.14, 0.7, 0.8])
        axCBarSigma = figSigma.add_axes([0.85, 0.1, 0.05, 0.8])

        for row in range(len(meanMatrix)):
            axMean.plot(pixelDAC, meanMatrix[row], color=self.getColor('viridis', len(meanMatrix), row))
            axSigma.plot(pixelDAC, sigmaMatrix[row], color=self.getColor('viridis', len(meanMatrix), row))

        self.getColorBar(axCBarMean, minIpixeldac*100, maxIpixeldac*100, NIpixeldac, label=r'$I_{\mathrm{pixelDAC}}$ (%)', rotation=0)
        self.getColorBar(axCBarSigma, minIpixeldac*100, maxIpixeldac*100, NIpixeldac, label=r'$I_{\mathrm{pixelDAC}}$ (%)', rotation=0)

        # Labels
        axMean.set_xlabel('PixelDAC')
        axMean.set_ylabel(r'$\mu_{\mathrm{THL}}$ (THL)')

        axSigma.set_xlabel('PixelDAC')
        axSigma.set_ylabel(r'$\sigma_{\mathrm{THL}}$ (THL)')

        figMean.show()
        figSigma.show()

        # Init plot
        # fig = plt.figure(figsize=(7, 4))
        # ax = fig.add_axes([0.1, 0.14, 0.7, 0.8])
        # Add colorbar later
        # axCBar = fig.add_axes([0.85, 0.1, 0.05, 0.8])

        # ax.hist(gaussCorrDict[pixelDAC], label=str(I_pixeldac), bins=np.arange(0, 1500, 5), alpha=0.5, color=self.getColor('viridis', len(I_pixeldacList), idx))

        # Add colorbar
        # self.getColorBar(axCBar, I_pixeldacList[0]*100, I_pixeldacList[-1]*100, N=len(I_pixeldacList), label=r'I_pixelDAC (\%)')

        # Standard deviation plot
        # figStd, axStd = plt.subplots()
        # axStd.plot(I_pixeldacList, stdList)

        # figStd.show()
        raw_input('')

        return

    def meanSigmaMatrixPlot(self, fn, pIpixeldac, pPixelDAC):
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        d = json.load( open(fn, 'rb') )
        meanMatrix = d['mean']
        print(meanMatrix)
        sigmaMatrix = d['sigma']

        figMean, axMean = plt.subplots()
        figSigma, axSigma = plt.subplots()

        # Ipixeldac params
        minIpixeldac, maxIpixeldac, NIpixeldac = pIpixeldac
        # pixelDAC params
        minDAC, maxDAC, NDAC = pPixelDAC

        # Calculate bin widths
        wIpixeldac = float(maxIpixeldac - minIpixeldac) / NIpixeldac
        wDAC = float(maxDAC - minDAC) / NDAC

        # Sigma plot
        im = axSigma.imshow(sigmaMatrix.T, extent=[minIpixeldac - 0.5*wIpixeldac, maxIpixeldac + 0.5*wIpixeldac, minDAC - 0.5*wDAC, maxDAC + 0.5*wDAC], aspect='auto')
        divider = make_axes_locatable(axSigma)
        caxSigma = divider.append_axes('right', size='5%', pad=0.1)
        figSigma.colorbar(im, cax=caxSigma)
        axSigma.set_xlabel('I_pixeldac'), axSigma.set_ylabel('pixelDAC')
        figSigma.show()

        # Mean plot
        im = axMean.imshow(meanMatrix.T, extent=[minIpixeldac - 0.5*wIpixeldac, maxIpixeldac + 0.5*wIpixeldac, minDAC - 0.5*wDAC, maxDAC + 0.5*wDAC], aspect='auto')
        divider = make_axes_locatable(axMean)
        caxMean = divider.append_axes('right', size='5%', pad=0.1)
        figMean.colorbar(im, cax=caxMean)
        axMean.set_xlabel('I_pixeldac'), axMean.set_ylabel('pixelDAC')
        figMean.show()

        raw_input('')

    def isLarge(self, pixel):
        if ( (pixel - 1) % 16 == 0 ) or ( pixel % 16 == 0 ) or ( (pixel + 1) % 16 == 0 ) or ( (pixel + 2) % 16 == 0 ):
            return False
        return True

    def energyToToTAtan(self, x, a, b, c, t):
        return np.where(x > b, a*(x - b) + c*np.arctan((x - b)/t), np.nan)

    def energyToToT(self, x, a, b, c, t, atan=False):
        if atan:
            return EnergyToToTAtan(x, a, b, c, t)
        else:
            return np.where(x > t, a*x + b + float(c)/(x - t), np.nan)

    def convertToDecimal(self, res, length=4):
        # Convert 4 hexadecimal characters each to int
        hexNumList = np.asarray( [int(hexNum, 16) for hexNum in textwrap.wrap(res.decode("utf-8"), length)] )
        return hexNumList.reshape((16, 16))

    # Infinite for loop
    def infinite_for(self):
        x = 0
        while True:
            yield x
            x += 1

    # Megalix control
    def megalix_connect(self, port):
        import serial
        return serial.Serial(port, stopbits=2)

    def megalix_set_kvpmA(self, mlx, voltage, current, dt=1):
        import time
        focus_size = 's'
        cmd = 'ei %s %s %d %s\n' % (voltage, current, dt, focus_size)
        mlx.write(cmd.encode('utf-8'))
        print('Sent command %s to megalix' % cmd)
        time.sleep( 1 )

    def megalix_xray_on(self, mlx):
        mlx.write('PREP 1\n'.encode('utf-8'))
        mlx.write('SWRON\n'.encode('utf-8'))

    def megalix_xray_off(self, mlx):
        mlx.write('SWROFF\n'.encode('utf-8'))
        mlx.write('PREP 0\n'.encode('utf-8'))




#!/usr/bin/env python
import dpx_control

PORT = '/dev/ttyUSB0'
CONFIG_FN = 'DPXConfig_22_6_109.conf'
CONFIG_DIR = 'config/'
CHIP_NUMS = [22, 6, 109]
IKRUM = [10, 30, 50] 
THL_SHIFT = 10

def main():
    # Establish connection
    thl_calib_files = None # [CONFIG_DIR + '/THLCalib_%d.hck' % CHIP for CHIP in CHIP_NUMS] 
    dpx = dpx_control.Dosepix(PORT, 2e6, CONFIG_DIR + '/' + CONFIG_FN, thl_calib_files=thl_calib_files)

    # Change Ikrum values
    new_peripherys = []
    for chip_idx in range(3):
        d = dpx.splitPerihperyDACs(dpx.peripherys + dpx.THLs[chip_idx], perc=False)
        d['I_krum'] = IKRUM[chip_idx]
        code = dpx.periheryDACsDictToCode(d, perc=False)
        code = code[:-4] + '%04x' % (int(dpx.THLs[chip_idx], 16) - THL_SHIFT)
        dpx.THLs[chip_idx] = '%04x' % (int(dpx.THLs[chip_idx], 16) - THL_SHIFT)

        dpx.peripherys = code[:-4]
        new_peripherys.append( code[:-4] )
        dpx.DPXWritePeripheryDACCommand(chip_idx + 1, code)
        print dpx.DPXReadPeripheryDACCommand(chip_idx + 1)
        print dpx.DPXReadOMRCommand(chip_idx + 1)

    for slot in range(1, 4):
        print dpx.DPXReadPeripheryDACCommand(slot)

    # THL scan
    THL = int(dpx.THLs[0], 16)
    THLlow = THL - 500
    THLhigh = THL + 500
    # dpx.energySpectrumTHL(slot=3, THLhigh=THLhigh, THLlow=THLlow, THLstep=1, timestep=0, intPlot=False)
    dpx.findNoise(slot=1, THLlow=THLlow, THLhigh=THLhigh, THLstep=1, timestep=0)
    return

    # XRT
    THLlow = 5700 # THL
    THLhigh = 6500 # 8000 # THLlow + 1000
    for voltage in [40]:
        for current in [20, 25, 30, 40, 50]:
            print 'Processing %d kV at %d mA' % (voltage, current)
            dpx.findNoise(slot=1, THLlow=THLlow, THLhigh=THLhigh, THLstep=10, timestep=0, outFn='noise_%d_%d.hck' % (voltage, current), megalix_port='/dev/ttyUSB1', megalix_settings=(voltage, current))
            raw_input('Done! Press any key to continue')
            print

    # Close connection
    dpx.close()

if __name__ == '__main__':
    main()




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(name='dpx_control',
    version='0.3.2',
    description='DPX control software',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Sebastian Schmidt',
    author_email='schm.seb@gmail.com',
    url="https://github.com/dosepix/dpx_control",
    project_urls={
        "Bug Tracker": "https://github.com/dosepix/dpx_control/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    license='GNU GPLv3',
    entry_points={
        'console_scripts' : [
            'dpx_control = dpx_control.dpx_control:main',
        ]
    },
    packages=["dpx_control"],
    install_requires=[
        'matplotlib',
        'numpy',
        'scipy',
        'pyserial',
        'configparser',
        'tqdm'
    ])




#!/usr/bin/env python
from __future__ import print_function
import dpx_control as dpx
import sys
import zerorpc
import numpy as np
import json
import pandas as pd
import time
import gevent
import serial.tools.list_ports

def main():
    addr = 'tcp://127.0.0.1:' + parse_port()
    s = zerorpc.Server(DPXapi())
    s.bind(addr)
    print('start running on {}'.format(addr))
    s.run()

def parse_port():
    port = 4242
    try:
        port = int(sys.argv[1])
    except Exception as e:
        pass
    return '{}'.format(port)

class DPXapi(object):
    # Function to check if server is ready
    def echo(self, text):
        """echo any text"""
        return text

    def connectDPX(self):
        self.dpxObj = dpx.Dosepix('/dev/ttyUSB0', 2e6, 'DPXConfig.conf')
        self.dpxObj.setGUI()
        return str('Success!')

    def disconnectDPX(self):
        del self.dpxObj


    def findDPX(self):
        ports = list(serial.tools.list_ports.comports())
        for p in ports:
            if 'FT232R USB UART' in p:
                return True
        else:
            return False

    def test(self):
        # return 'Test'
        return [1, 2, 3]

    @zerorpc.stream
    def measureToT(self, out, cnt):
        # Function yields histogram data
        rand = []
        bins = np.arange(100)

        cnt = 100

        for f in np.arange(int(cnt)):
            gevent.sleep(0)
            d = np.random.normal(30, 10, 1000)
            rand += list( d )
            h, b = np.histogram(rand, bins=bins)

            data = json.dumps({'frame': f / 100., 'data': [{'ToT': b[:-1][i], 'Counts': h[i]} for i in range(len(h))]})
            yield data
            time.sleep(0.1)
        # return str('Not implemented!')
        return

        # Settings
        slot = 1
        cnt = int( cnt )    # Ensure cnt is given as int
        # cnt = 10000
        # out = 'ToTMeasurement/'

        # Get generator object
        ToTgen = self.dpxObj.measureToT(slot=slot, outDir=out, cnt=cnt, storeEmpty=False, logTemp=False, paramsDict=None, intPlot=False)
        for b, h in ToTgen:
            gevent.sleep(0)     # Required for heartbeats
            data = json.dumps({'frame': 1., 'data': [{'ToT': b[:-1][i], 'Counts': h[i]} for i in range(len(h))]})
            yield data
        return

    @zerorpc.stream
    def testPlot(self):
        # Processing example

        '''
        for x in np.arange(100):
            # Do some calculations
            yield pd.DataFrame({'status': x}).to_json()
        '''

        xStart = np.arange(10000)
        for x_ in xStart:
            gevent.sleep(0)     # Required for heartbeats
            x = np.linspace(x_, x_+10, 100).tolist()
            y = np.sin(x).tolist()
            # data = pd.DataFrame({'status': x_, 'data': {'x': x.tolist(), 'y': y.tolist()}})
            data = json.dumps({'status': x_ / 10000., 'data': [{'x': x[i], 'y': y[i]} for i in range(len(x))]})
            # print( data )
            yield data # pd.DataFrame({'x': x.tolist(), 'y': y.tolist()}).to_json(orient='records')
            time.sleep(.1)

if __name__ == '__main__':
    main()




from __future__ import print_function

import serial
import os
import os.path
import numpy as np
import json
import configparser

try:
  basestring
except NameError:
  basestring = str

class Config(object):
    def getConfigDPX(self, portName, baudRate, configFn, Ikrum=None, eye_lens=False):
        self.ser = serial.Serial(self.portName, self.baudRate)
        assert self.ser.is_open, 'Error: Could not establish serial connection!'

        # Read config
        if self.eye_lens:
            slot_fac = 1
        else:
            slot_fac = 3
        
        # Standard values
        self.peripherys = ['dc310bc864508230768080ff0064'] * slot_fac
        self.OMR = ['39ffc0'] * slot_fac
        self.THLs = ['153b'] * slot_fac
        self.confBits = ['0' * 512] * slot_fac
        self.pixelDAC = ['0' * 512] * slot_fac
        self.binEdges = ['0' * 1024] * slot_fac

        # Set standard Ikrum
        if Ikrum is not None:
            for p_idx, p in enumerate( self.peripherys ):
                d = self.splitPerihperyDACs(p + self.THLs[p_idx], perc=False, show=False)
                d['I_krum'] = Ikrum[p_idx]
                code = self.periheryDACsDictToCode(d, perc=False)
                self.peripherys[p_idx] = code[:-4]

        if self.configFn is None:
            print('Config file not found. Please run THL equalization first. Using standard values.')
        else:
            if isinstance(self.configFn, basestring):
                if not os.path.isfile(conf):
                    print('Config file not found. Please run THL equalization first. Using standard values.')
                else:
                    print('Only one config file specified. Using same file for all slots.')
                    for slot in self.slot_range:
                        self.readConfig(self.configFn, slot=slot)
            else:
                for idx, conf in enumerate(self.configFn):
                    if conf is not None:
                        if os.path.isfile(conf):
                            self.readConfig(conf, slot=idx+1)
                        else:
                            print('Config file not found. Please run THL equalization first. Using standard values.')

        self.initDPX()

        # If no THL calibration file is present
        if self.thl_calib_files is None:
            print('Warning: No THL calibration file set! Functions using THL edges won\'t be usuable')
            return

        # Load THL calibration data
        for thl_calib_file in self.thl_calib_files:
            if thl_calib_file is None or not os.path.isfile(thl_calib_file):
                print('Need the specified THL Calibration file %s!' % thl_calib_file)
                self.THLEdgesLow.append( None ) 
                self.THLEdgesHigh.append( None )
                self.THLEdges.append( None )
                self.THLFitParams.append( None )
            else:
                if thl_calib_file.endswith('.json'):
                    # JSON
                    print(thl_calib_file)
                    with open(thl_calib_file, 'r') as f:
                        d = json.load(f)
                else:
                    print('Cannot open THL calibration file with extension %s' % thl_calib_file.split('.')[-1])
                    self.THLEdgesLow.append( None ) 
                    self.THLEdgesHigh.append( None )
                    self.THLEdges.append( None )
                    self.THLFitParams.append( None )

                self.load_THLEdges(d)
    
    def load_THLEdges(self, d):
        self.voltCalib.append( np.asarray(d['Volt']) / max(d['Volt']) )
        self.THLCalib.append( np.asarray(d['ADC']) )

        THLLow, THLHigh, THLFitParams = self.THLCalibToEdges(d)

        # Rounding
        THLLow = np.ceil(THLLow)
        THLHigh = np.floor(THLHigh)

        self.THLEdgesLow.append(THLLow), self.THLEdgesHigh.append(THLHigh), self.THLFitParams.append(THLFitParams)
        print(self.THLEdgesLow[-1], self.THLEdgesHigh[-1])
        
        # Combine
        THLEdges = []
        for i in range(len(self.THLEdgesLow[-1])):
            THLEdges += list( np.arange(self.THLEdgesLow[-1][i], self.THLEdgesHigh[-1][i] + 1) )
        self.THLEdges.append( THLEdges ) 

    def splitPerihperyDACs(self, code, perc=False, show=False):
        if perc:
            percEightBit = float( 2**8 )
            percNineBit = float( 2**9 )
            percThirteenBit = float( 2**13 )
        else:
            percEightBit, percNineBit, percThirteenBit = 1, 1, 1

        if show:
            print(code)
        code = format(int(code, 16), 'b')
        if show:
            print(code)
        d = {'V_ThA': int(code[115:], 2) / percThirteenBit,
            'V_tpref_fine': int(code[103:112], 2) / percNineBit,
            'V_tpref_coarse': int(code[88:96], 2) / percEightBit, 
            'I_tpbufout': int(code[80:88], 2) / percEightBit, 
            'I_tpbufin': int(code[72:80], 2) / percEightBit, 
            'I_disc2': int(code[64:72], 2) / percEightBit, 
            'I_disc1': int(code[56:64], 2) / percEightBit, 
            'V_casc_preamp': int(code[48:56], 2) / percEightBit, 
            'V_gnd': int(code[40:48], 2) / percEightBit, 
            'I_preamp': int(code[32:40], 2) / percEightBit, 
            'V_fbk': int(code[24:32], 2) / percEightBit, 
            'I_krum': int(code[16:24], 2) / percEightBit, 
            'I_pixeldac': int(code[8:16], 2) / percEightBit, 
            'V_casc_reset': int(code[:8], 2) / percEightBit}

        if show:
            print('PeripheryDAC values in', end='')
            if perc:
                print('percent:')
            else:
                print('DAC:')

        return d

    def periheryDACsDictToCode(self, d, perc=False):
        if perc:
            percEightBit = float( 2**8 )
            percNineBit = float( 2**9 )
            percThirteenBit = float( 2**13 )
        else:
            percEightBit, percNineBit, percThirteenBit = 1, 1, 1

        code = 0
        code |= int(d['V_ThA'] * percThirteenBit) 
        code |= (int(d['V_tpref_fine'] * percNineBit) << 25 - 9)
        code |= (int(d['V_tpref_coarse'] * percEightBit) << 40 - 8)
        code |= (int(d['I_tpbufout'] * percEightBit) << 48 - 8)
        code |= (int(d['I_tpbufin'] * percEightBit) << 56 - 8)
        code |= (int(d['I_disc2'] * percEightBit) << 64 - 8)
        code |= (int(d['I_disc1'] * percEightBit) << 72 - 8)
        code |= (int(d['V_casc_preamp'] * percEightBit) << 80 - 8)
        code |= (int(d['V_gnd'] * percEightBit) << 88 - 8)
        code |= (int(d['I_preamp'] * percEightBit) << 96 - 8)
        code |= (int(d['V_fbk'] * percEightBit) << 104 - 8)
        code |= (int(d['I_krum'] * percEightBit) << 112 - 8)
        code |= (int(d['I_pixeldac'] * percEightBit) << 120 - 8)
        code |= (int(d['V_casc_reset'] * percEightBit) << 128 - 8)

        return '%04x' % code

    def readConfig(self, configFn, slot=1):
        config = configparser.ConfigParser()
        config.read(configFn)

        # Mandatory sections
        sectionList = ['Peripherydac', 'OMR', 'Equalisation']

        # Check if set, else throw error
        for section in config.sections():
            assert section in sectionList, 'Config: %s is a mandatory section and has to be specified' % section

        # Read Peripherydac
        if 'code' in config['Peripherydac']:
            self.peripherys = config['Peripherydac']
        else:
            PeripherydacDict = {}
            PeripherydacCodeList = ['V_ThA', 'V_tpref_fine', 'V_tpref_coarse', 'I_tpbufout', 'I_tpbufin', 'I_disc1', 'I_disc2', 'V_casc_preamp', 'V_gnd', 'I_preamp', 'V_fbk', 'I_krum', 'I_pixeldac', 'V_casc_reset']
            for PeripherydacCode in PeripherydacCodeList:
                assert PeripherydacCode in config['Peripherydac'], 'Config: %s has to be specified in OMR section!' % PeripherydacCode

                PeripherydacDict[PeripherydacCode] = int(float( config['Peripherydac'][PeripherydacCode] ))
            self.peripherys[slot - 1] = self.periheryDACsDictToCode(PeripherydacDict)[:-4]
            self.THLs[slot - 1] = '%04x' % PeripherydacDict['V_ThA']

        # Read OMR
        if 'code' in config['OMR']:
            self.OMR[slot - 1] = config['OMR']['code']
        else:
            OMRList = []

            OMRCodeList = ['OperationMode', 'GlobalShutter', 'PLL', 'Polarity', 'AnalogOutSel', 'AnalogInSel', 'OMRDisableColClkGate']
            for OMRCode in OMRCodeList:
                assert OMRCode in config['OMR'], 'Config: %s has to be specified in OMR section!' % OMRCode

                OMRList.append(config['OMR'][OMRCode])

            self.OMR[slot - 1] = OMRList

        # Equalisation
        # confBits - optional field
        if 'confBits' in config['Equalisation']:
            self.confBits[slot - 1] = config['Equalisation']['confBits']
        else:
            # Use all pixels
            self.confBits[slot - 1] = '00' * 256

        # pixelDAC
        assert 'pixelDAC' in config['Equalisation'], 'Config: pixelDAC has to be specified in Equalisation section!'
        self.pixelDAC[slot - 1] = config['Equalisation']['pixelDAC']

        # binEdges
        assert 'binEdges' in config['Equalisation'], 'Config: binEdges has to be specified in Equalisation section!'
        self.binEdges[slot - 1] = config['Equalisation']['binEdges']

        return

    def writeConfig(self, configFn, slot=1):
        config = configparser.ConfigParser()
        d = self.splitPerihperyDACs(self.peripherys[slot-1] + self.THLs[slot-1])
        d['V_ThA'] = int(self.THLs[slot-1], 16)
        config['Peripherydac'] = d

        if not isinstance(self.OMR[slot-1], basestring):
            OMRCodeList = ['OperationMode', 'GlobalShutter', 'PLL', 'Polarity', 'AnalogOutSel', 'AnalogInSel', 'OMRDisableColClkGate']
            config['OMR'] = {OMRCode: self.OMR[slot-1][i] for i, OMRCode in enumerate(OMRCodeList)}
        else:
            config['OMR'] = {'code': self.OMR[slot-1]}

        config['Equalisation'] = {'pixelDAC': self.pixelDAC[slot-1], 'confBits': self.confBits[slot-1], 'binEdges': ''.join(self.binEdges['Slot%d' % slot])}

        with open(configFn, 'w') as configFile:
            config.write(configFile)




from __future__ import print_function

from collections import namedtuple
# === CONSTANTS ===
_startOfTransmission   = b"\x02"
_endOfTransmission     = b"\x03"

# = Receiver =
_receiverDPX1 = '01'   # Dosepix
_receiverDPX2 = '02'
_receiverDPX3 = '03'
_receiverMC = '04'     # Microcontroller
_receiverVC = '09'     # Voltage Controller
_receiverHV = '11'     # High Voltage

# = Subreceiber =
_subReceiverNone = '000'

# = Sender =
_senderPC = '10'

# = Commands =
_commandNone = ''
_commandNoneLength = '0' * 6

# = HV Commands =
_HVenable = '000'
_HVdisable = '001'
_HVisEnabled = '002'
_HVsetDAC = '003'
_HVgetDAC = '004'

# = VC Commands =
_VCset3V3 = '000'
_VCset1V8 = '001'
_VCgetVoltage = '002'

# = MC Commands =
_MCgetVersion = '000'
_MCLEDenable = '001'
_MCLEDdisable = '002'
_MCgetADCvalue = '003'
_MCsetSPIclock1 = '004'
_MCsetSPIclock2 = '005'
_MCsetSPIclock3 = '006'

# = DPX Commands =
_DPXwriteOMRCommand = '001' 
_DPXwriteConfigurationCommand = '002'
_DPXwriteSingleThresholdCommand = '003'
_DPXwritePixelDACCommand = '004'
_DPXwritePeripheryDACCommand = '005'
_DPXwriteColSelCommand = '006'
_DPXburnSingleFuseCommand = '007'
_DPXreadOMRCommand = '008'
_DPXreadConfigurationCommand = '009'
_DPXreadDigitalThresholdsCommand = '010'
_DPXreadPixelDACCommand = '011'
_DPXreadPeripheryDACCommand = '012'
_DPXreadColumnTestPulseCommand = '013'
_DPXreadColSelCommand = '014'
_DPXreadChipIdCommand = '015'

_DPXglobalResetCommand = '020'
_DPXdataResetCommand = '021'

_DPXreadToTDataDosiModeCommand = '050'
_DPXreadBinDataDosiModeCommand = '051'
_DPXreadToTDatakVpModeCommand = '053'
_DPXreadToTDataIntegrationModeCommand = '054'

_DPXgeneralTestPulse = '057'
_DPXreadToTDataInkVpModeWithFixedFrameSizeCommand = '066'
_DPXgeneralMultiTestPulse = '067'

_DPXstartStreamingReadout = '068'
_DPXstopStreamingReadout ='069'

# = Custom Functions =
_DPXreadToTDataDosiModeMultiCommand = '090'

# = CRC =
_CRC = 'FFFF'

# = OMR =
OMROperationModeType = namedtuple("OMROperationMode", "DosiMode TestWakeUp PCMode IntegrationMode")
_OMROperationMode = OMROperationModeType(
    DosiMode = 0b00 << 22,
    TestWakeUp = 0b01 << 22,
    PCMode = 0b10 << 22,
    IntegrationMode = 0b11 << 22)

OMRGlobalShutterType = namedtuple("OMRGlobalShutter", "ClosedShutter OpenShutter")
_OMRGlobalShutter = OMRGlobalShutterType(
    ClosedShutter = 0b0 << 21,
    OpenShutter = 0b1 << 21)

# Do not use 200 MHz!
OMRPLLType = namedtuple("OMRPLL", "Direct f16_6MHz f20MHz f25MHz f33_2MHz f50MHz f100MHz")
_OMRPLL = OMRPLLType(Direct = 0b000 << 18,
    f16_6MHz = 0b001 << 18,
    f20MHz = 0b010 << 18,
    f25MHz = 0b011 << 18,
    f33_2MHz = 0b100 << 18,
    f50MHz = 0b101 << 18,
    f100MHz = 0b110 << 18)

OMRPolarityType = namedtuple("OMRPolarity", "electron hole")
_OMRPolarity = OMRPolarityType(
    hole = 0b0 << 17,
    electron = 0b1 << 17)

OMRAnalogOutSelType = namedtuple("OMRAnalogOutSel", "V_ThA V_TPref_fine V_casc_preamp V_fbk V_TPref_coarse V_gnd I_preamp I_disc1 I_disc2 V_TPbufout V_TPbufin I_krum I_dac_pixel V_bandgap V_casc_krum Temperature V_per_bias V_cascode_bias High_Z")
_OMRAnalogOutSel = OMRAnalogOutSelType(
    V_ThA = 0b00001 << 12,
    V_TPref_fine = 0b00010 << 12, 
    V_casc_preamp = 0b00011 << 12,
    V_fbk = 0b00100 << 12,
    V_TPref_coarse = 0b00101 << 12,
    V_gnd = 0b00110 << 12,
    I_preamp = 0b00111 << 12,
    I_disc1 = 0b01000 << 12,
    I_disc2 = 0b01001 << 12,
    V_TPbufout = 0b01010 << 12,
    V_TPbufin = 0b01011 << 12,
    I_krum = 0b01100 << 12,
    I_dac_pixel = 0b01101 << 12,
    V_bandgap = 0b01110 << 12,
    V_casc_krum = 0b01111 << 12,
    Temperature = 0b11011 << 12,
    V_per_bias = 0b11100 << 12,
    V_cascode_bias = 0b11101 << 12,
    High_Z = 0b11111 << 12)

OMRAnalogInSelType = namedtuple("OMRAnalogInSel", "V_ThA V_TPref_fine V_casc_preamp V_fbk V_TPref_coarse V_gnd I_preamp I_disc1 I_disc2 V_TPbufout V_TPbufin I_krum I_dac_pixel V_bandgap V_casc_krum Temperature V_per_bias V_cascode_bias V_no")
_OMRAnalogInSel = OMRAnalogInSelType(
    V_ThA = 0b00001 << 7,
    V_TPref_fine = 0b00010 << 7, 
    V_casc_preamp = 0b00011 << 7,
    V_fbk = 0b00100 << 7,
    V_TPref_coarse = 0b00101 << 7,
    V_gnd = 0b00110 << 7,
    I_preamp = 0b00111 << 7,
    I_disc1 = 0b01000 << 7,
    I_disc2 = 0b01001 << 7,
    V_TPbufout = 0b01010 << 7,
    V_TPbufin = 0b01011 << 7,
    I_krum = 0b01100 << 7,
    I_dac_pixel = 0b01101 << 7,
    V_bandgap = 0b01110 << 7,
    V_casc_krum = 0b01111 << 7,
    Temperature = 0b11011 << 7,
    V_per_bias = 0b11100 << 7,
    V_cascode_bias = 0b11101 << 7,
    V_no = 0b11111 << 7)

OMRDisableColClkGateType = namedtuple("OMRDisableColClkGate", "Enabled Disabled")
_OMRDisableColClkGate = OMRDisableColClkGateType(
    Enabled = 0b0 << 6,
    Disabled = 0b1 << 6)

# = ConfBits =
ConfBitsType = namedtuple("ConfBits", "MaskBit TestBit_Analog TestBit_Digital")
_ConfBits = ConfBitsType(
    MaskBit = 0b1 << 2,
    TestBit_Analog = 0b1 << 1,
    TestBit_Digital = 0b1 << 0)
